from threading import Event


class ImmediateScheduler(object):
    def call(self, fn):
        try:
            fn()
        except:
            pass

    def wait(self, promise, timeout=None):
        e = Event()

        def on_resolve_or_reject(_):
            e.set()

        promise._then(on_resolve_or_reject, on_resolve_or_reject)
        waited = e.wait(timeout)
        if not waited:
            raise Exception("Timeout")



import functools
import inspect
import types
import warnings
import sys


def warn(msg):
    warnings.simplefilter('always', DeprecationWarning)  # turn off filter
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
    warnings.simplefilter('default', DeprecationWarning)  # reset filter


class deprecated(object):

    def __init__(self, reason, name=None):
        if inspect.isclass(reason) or inspect.isfunction(reason):
            raise TypeError("Reason for deprecation must be supplied")
        self.reason = reason
        self.name = name

    def __call__(self, cls_or_func):
        if inspect.isfunction(cls_or_func):
            fmt = "Call to deprecated function or method {name} ({reason})."

        elif inspect.isclass(cls_or_func):
            fmt = "Call to deprecated class {name} ({reason})."

        else:
            raise TypeError(type(cls_or_func))

        msg = fmt.format(name=self.name or cls_or_func.__name__, reason=self.reason)

        @functools.wraps(cls_or_func)
        def new_func(*args, **kwargs):
            warn(msg)
            return cls_or_func(*args, **kwargs)

        return new_func


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,  # type: tuple
    integer_types = int,  # type: tuple
    class_types = type,  # type: tuple
    text_type = str
    binary_type = bytes
else:
    string_types = basestring,  # type: tuple
    integer_types = (int, long)  # type: tuple
    class_types = (type, types.ClassType)  # type: tuple
    text_type = unicode
    binary_type = str



from .pyutils.version import get_version


try:
    # This variable is injected in the __builtins__ by the build
    # process. It used to enable importing subpackages when
    # the required packages are not installed
    __SETUP__  # type: ignore
except NameError:
    __SETUP__ = False


VERSION = (2, 1, 0, 'alpha', 0)

__version__ = get_version(VERSION)

if not __SETUP__:
    from .promise import (
        Promise,
        promise_for_dict,
        promisify,
        is_thenable,
        async_instance,
        get_default_scheduler,
        set_default_scheduler
    )
    from .schedulers.immediate import ImmediateScheduler

    __all__ = [
        'Promise',
        'promise_for_dict',
        'promisify',
        'is_thenable',
        'async_instance',
        'get_default_scheduler',
        'set_default_scheduler',
        'ImmediateScheduler'
    ]



import sys
from setuptools import setup, find_packages

if sys.version_info[0] < 3:
    import __builtin__ as builtins
else:
    import builtins

builtins.__SETUP__ = True

version = __import__('promise').get_version()


IS_PY3 = sys.hexversion >= 0x03000000

tests_require = [
    'pytest>=2.7.3', 'pytest-cov', 'coveralls', 'futures', 'pytest-benchmark',
    'mock',
]
if IS_PY3:
    tests_require += ['pytest-asyncio']


setup(
    name='promise',
    version=version,
    description='Promises/A+ implementation for Python',
    long_description=open('README.rst').read(),
    url='https://github.com/syrusakbary/promise',
    download_url='https://github.com/syrusakbary/promise/releases',
    author='Syrus Akbary',
    author_email='me@syrusakbary.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: PyPy',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='concurrent future deferred promise',
    packages=find_packages(exclude=['tests']),
    extras_require={
        'test': tests_require,
    },
    install_requires=[
        'typing', 'six'
    ],
    tests_require=tests_require, )



from functools import partial
from collections import Iterable


class PromiseList(object):

    __slots__ = ('_values', '_length', '_total_resolved', 'promise', '_promise_class')

    def __init__(self, values, promise_class):
        self._promise_class = promise_class
        self.promise = self._promise_class()

        self._values = values
        self._length = 0
        self._total_resolved = 0
        self._init()

    def __len__(self):
        return self._length

    def _init(self):
        Promise = self._promise_class
        values = self._values
        if Promise.is_thenable(values):
            values = Promise._try_convert_to_promise(self._values)._target()
            if values.is_fulfilled:
                values = values._value()
            elif values.is_rejected:
                return self._reject(values._reason())
            # Is pending
            else:
                self.promise._is_async_guaranteed = True
                return values._then(
                    self._init,
                    self._reject,
                )

        if not isinstance(values, Iterable):
            err = Exception("PromiseList requires an iterable. Received {}.".format(repr(values)))
            self.promise._reject_callback(err, False)
            return

        if not values:
            self._resolve([])
            return

        self._iterate(values)

    def _iterate(self, values):
        Promise = self._promise_class
        is_resolved = False
        self._length = len(values)
        self._values = [None] * self._length

        result = self.promise

        for i, val in enumerate(values):
            if Promise.is_thenable(val):
                maybe_promise = Promise._try_convert_to_promise(val)._target()
                # if is_resolved:
                #     # maybe_promise.suppressUnhandledRejections
                #     pass
                if maybe_promise.is_pending:
                    maybe_promise._add_callbacks(
                        partial(self._promise_fulfilled, i=i),
                        self._promise_rejected,
                        None
                    )
                    self._values[i] = maybe_promise
                elif maybe_promise.is_fulfilled:
                    is_resolved = self._promise_fulfilled(maybe_promise._value(), i)
                elif maybe_promise.is_rejected:
                    is_resolved = self._promise_rejected(maybe_promise._reason())

            else:
                is_resolved = self._promise_fulfilled(val, i)

            if is_resolved:
                break

        if not is_resolved:
            result._is_async_guaranteed = True

    def _promise_fulfilled(self, value, i):
        if self.is_resolved:
            return
        # assert not self.is_resolved
        # assert isinstance(self._values, Iterable)
        # assert isinstance(i, int)
        self._values[i] = value
        self._total_resolved += 1
        if self._total_resolved >= self._length:
            self._resolve(self._values)
            return True
        return False

    def _promise_rejected(self, reason):
        if self.is_resolved:
            return
        # assert not self.is_resolved
        # assert isinstance(self._values, Iterable)
        self._total_resolved += 1
        self._reject(reason)
        return True

    @property
    def is_resolved(self):
        return self._values is None

    def _resolve(self, value):
        assert not self.is_resolved
        assert not isinstance(value, self._promise_class)
        self._values = None
        self.promise._fulfill(value)

    def _reject(self, reason):
        assert not self.is_resolved
        self._values = None
        self.promise._reject_callback(reason, False)
